/*
 *  GuiLoader/C++, a C++ binding to GuiLoader
 *  Copyright (c) 2006-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "guiloader-c++_impl.h"

namespace Gui {

// Enter -----------------------------------------------------------------------

GUILOADER_CPP_API Enter::Enter() {
	guiloader_enter();
	LoaderInit::enter();
}

GUILOADER_CPP_API Enter::~Enter() {
	LoaderInit::leave();
	guiloader_leave();
}

// Loader ----------------------------------------------------------------------

GUILOADER_CPP_API PLoader Loader::create() {
	Loader * loader= new Loader();
	return PLoader(loader);
}

GUILOADER_CPP_API void Loader::set_preview_mode(bool preview) {
	gui_loader_set_preview_mode(GUI_LOADER(gobj()), preview);
}

GUILOADER_CPP_API void Loader::set_take_objects(bool take) {
	gui_loader_set_take_objects(GUI_LOADER(gobj()), take);
}

GUILOADER_CPP_API void Loader::load_from_string(const Glib::ustring & data) {
	gui_loader_load_from_string(GUI_LOADER(gobj()), data.c_str(), data.bytes());
}

GUILOADER_CPP_API void Loader::load_from_file(const std::string & filename) {
	GError * error= NULL;
	gui_loader_load_from_file(GUI_LOADER(gobj()), filename.c_str(), &error);
	if(error)
		Glib::Error::throw_exception(error);
}

GUILOADER_CPP_API void Loader::clear() {
	gui_loader_clear(GUI_LOADER(gobj()));
}

GUILOADER_CPP_API void Loader::add_custom_constructor(const std::string & type, const CustomConstructor & slot) {
	CustomConstructor * constructor= new CustomConstructor(slot);

	struct temp {

		static GtkWidget * create(const char * type, const char * meta, const char * text, gpointer data) {
			try {
				CustomConstructor & slot= *static_cast<CustomConstructor *>(data);
				Gtk::Widget * widget= slot(type, meta, text);
				CHECK(widget && !dynamic_cast<Gtk::Window *>(widget));
				widget->set_manage();
				return widget->gobj();
			} catch(...) {
				Glib::exception_handlers_invoke();
				return NULL;
			}
		}

		static void notify(gpointer data) {
			CustomConstructor * slot= static_cast<CustomConstructor *>(data);
			delete slot;
		}

	};

	gui_loader_add_custom_constructor(
		GUI_LOADER(gobj()),
		type.c_str(),
		&temp::create,
		constructor,
		&temp::notify
	);

}

GUILOADER_CPP_API Loader::Connectors Loader::query_connectors(const std::string & name, int count) {
	GQueue * queue= gui_loader_query_connectors(GUI_LOADER(gobj()), name.c_str(), count);
	Connectors result;
	result.reserve(count);
	GList * elem;
	for(elem= g_queue_peek_head_link(queue); elem; elem= g_list_next(elem)) {
		GuiConnectInfo info;
		gui_loader_get_connect_info(GUI_LOADER(gobj()), static_cast<GuiConnector *>(elem->data), &info);
		result.resize(result.size()+1);
		Connector & connector= result.back();
		connector.object= Wrap(info.object);
		connector.signal= LoaderInit::get()->findSignal(info.signame, G_OBJECT_TYPE(info.object));
		connector.after= info.after;
	}
	return result;
}

GUILOADER_CPP_API Glib::Object * Loader::bind_object(const std::string & name) {
	return Wrap(G_OBJECT(gui_loader_bind_object(GUI_LOADER(gobj()), name.c_str())));
}

GUILOADER_CPP_API void Loader::error(int code) {
	switch(code) {
		case ecConnectFailed:
			FatalError(_("Connecting signals failed: internal error"));
			break;
		case ecConnectBadType:
			FatalError(_("Connecting signals failed: invalid slot type"));
			break;
		case ecBindBadType:
			FatalError(_("Binding variables failed: invalid object type"));
			break;
		default:
			FatalError(_("Unknown error"));
	};
}

Loader::Loader() :
	Glib::Object(G_OBJECT(gui_loader_new()))
{
}

// -----------------------------------------------------------------------------

void CheckFailed(const char * expr, const char * file, int line) {
	boost::format msg(_("Check failed: %s, file: %s, line: %d"));
	msg % expr % file % line;
	FatalError(msg.str());
}

void FatalError(const std::string & message) {
	g_critical(message.c_str());
	#ifdef GUILOADER_CPP_SANDBOX
		assert(0);
	#endif
	Gtk::MessageDialog dialog(
		_("Fatal Error") + Glib::ustring(" [" G_LOG_DOMAIN "]"), false,
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

Glib::Object * Wrap(GObject * gobj) {
	return dynamic_cast<Glib::Object *>(Glib::wrap_auto(gobj));
}

// LoaderInit ------------------------------------------------------------------

LoaderInit * LoaderInit::get() {
	CHECK(ptr());
	return ptr();
}

void LoaderInit::enter() {
	initGettext();
	if(!ptr())
		ptr()= new LoaderInit();
	else
		++ptr()->refcounter;
}

void LoaderInit::leave() {
	CHECK(ptr() && ptr()->refcounter > 0);
	--ptr()->refcounter;
	if(ptr()->refcounter == 0) {
		delete ptr();
		ptr()= NULL;
	}
}

Signal * LoaderInit::findSignal(const std::string & name, GType type) {
	Signals::iterator it= signals.lower_bound(name);
	if(it!=signals.end())
		do {
			if(g_type_is_a(type, it->second.first))
				return it->second.second;
			++it;
		} while(it!=signals.end() && it->first==name);
	boost::format msg(_("Signal '%1%' was not defined for %2%"));
	msg % name % g_type_name(type);
	FatalError(msg.str());
	return NULL;
}

LoaderInit * & LoaderInit::ptr() {
	static LoaderInit * loaderInit= NULL;
	return loaderInit;
}

void LoaderInit::initGettext() {
	#if ENABLE_NLS
		static bool done= false;
		if(!done) {
			bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
			bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
			done= true;
		}
	#endif
}

LoaderInit::LoaderInit() {

	CHECK(GUILOADER_VERSION/1000 <= guiloader_version()/1000);

	refcounter= 1;

	addingSignals(GTK_TYPE_WIDGET);
	addSignal("button-press-event", &Gtk::Widget::signal_button_press_event);
	addSignal("button-release-event", &Gtk::Widget::signal_button_release_event);
	addSignal("child-notify", &Gtk::Widget::signal_child_notify);
	addSignal("client-event", &Gtk::Widget::signal_client_event);
	addSignal("configure-event", &Gtk::Widget::signal_configure_event);
	addSignal("delete-event", &Gtk::Widget::signal_delete_event);
	addSignal("direction-changed", &Gtk::Widget::signal_direction_changed);
	addSignal("drag-begin", &Gtk::Widget::signal_drag_begin);
	addSignal("drag-data-delete", &Gtk::Widget::signal_drag_data_delete);
	addSignal("drag-data-get", &Gtk::Widget::signal_drag_data_get);
	addSignal("drag-data-received", &Gtk::Widget::signal_drag_data_received);
	addSignal("drag-drop", &Gtk::Widget::signal_drag_drop);
	addSignal("drag-end", &Gtk::Widget::signal_drag_end);
	addSignal("drag-leave", &Gtk::Widget::signal_drag_leave);
	addSignal("drag-motion", &Gtk::Widget::signal_drag_motion);
	addSignal("enter-notify-event", &Gtk::Widget::signal_enter_notify_event);
	addSignal("event", &Gtk::Widget::signal_event);
	addSignal("event-after", &Gtk::Widget::signal_event_after);
	addSignal("expose-event", &Gtk::Widget::signal_expose_event);
	addSignal("focus", &Gtk::Widget::signal_focus);
	addSignal("focus-in-event", &Gtk::Widget::signal_focus_in_event);
	addSignal("focus-out-event", &Gtk::Widget::signal_focus_out_event);
	addSignal("grab-focus", &Gtk::Widget::signal_grab_focus);
	addSignal("grab-notify", &Gtk::Widget::signal_grab_notify);
	addSignal("hide", &Gtk::Widget::signal_hide);
	addSignal("hierarchy-changed", &Gtk::Widget::signal_hierarchy_changed);
	addSignal("key-press-event", &Gtk::Widget::signal_key_press_event);
	addSignal("key-release-event", &Gtk::Widget::signal_key_release_event);
	addSignal("leave-notify-event", &Gtk::Widget::signal_leave_notify_event);
	addSignal("map", &Gtk::Widget::signal_map);
	addSignal("map-event", &Gtk::Widget::signal_map_event);
	addSignal("mnemonic-activate", &Gtk::Widget::signal_mnemonic_activate);
	addSignal("motion-notify-event", &Gtk::Widget::signal_motion_notify_event);
	addSignal("no-expose-event", &Gtk::Widget::signal_no_expose_event);
	addSignal("parent-set", &Gtk::Widget::signal_parent_changed);
	addSignal("property-notify-event", &Gtk::Widget::signal_property_notify_event);
	addSignal("proximity-in-event", &Gtk::Widget::signal_proximity_in_event);
	addSignal("proximity-out-event", &Gtk::Widget::signal_proximity_out_event);
	addSignal("realize", &Gtk::Widget::signal_realize);
	addSignal("screen-changed", &Gtk::Widget::signal_screen_changed);
	addSignal("scroll-event", &Gtk::Widget::signal_scroll_event);
	addSignal("selection-clear-event", &Gtk::Widget::signal_selection_clear_event);
	addSignal("selection-get", &Gtk::Widget::signal_selection_get);
	addSignal("selection-notify-event", &Gtk::Widget::signal_selection_notify_event);
	addSignal("selection-received", &Gtk::Widget::signal_selection_received);
	addSignal("selection-request-event", &Gtk::Widget::signal_selection_request_event);
	addSignal("show", &Gtk::Widget::signal_show);
	addSignal("size-allocate", &Gtk::Widget::signal_size_allocate);
	addSignal("size-request", &Gtk::Widget::signal_size_request);
	addSignal("state-changed", &Gtk::Widget::signal_state_changed);
	addSignal("style-set", &Gtk::Widget::signal_style_changed);
	addSignal("unmap", &Gtk::Widget::signal_unmap);
	addSignal("unmap-event", &Gtk::Widget::signal_unmap_event);
	addSignal("unrealize", &Gtk::Widget::signal_unrealize);
	addSignal("visibility-notify-event", &Gtk::Widget::signal_visibility_notify_event);
	addSignal("window-state-event", &Gtk::Widget::signal_window_state_event);
	addSignal("composited-changed", &Gtk::Widget::signal_composited_changed);
	addSignal("query-tooltip", &Gtk::Widget::signal_query_tooltip);
	addSignal("damage-event", &Gtk::Widget::signal_damage_event);
	addSignal("drag-failed", &Gtk::Widget::signal_drag_failed);
	addSignal("popup-menu", &Gtk::Widget::signal_popup_menu);

	addingSignals(GTK_TYPE_CONTAINER);
	addSignal("add", &Gtk::Container::signal_add);
	addSignal("check-resize", &Gtk::Container::signal_check_resize);
	addSignal("remove", &Gtk::Container::signal_remove);
	addSignal("set-focus-child", &Gtk::Container::signal_set_focus_child);

	addingSignals(GTK_TYPE_HANDLE_BOX);
	addSignal("child-attached", &Gtk::HandleBox::signal_child_attached);
	addSignal("child-detached", &Gtk::HandleBox::signal_child_detached);

	addingSignals(GTK_TYPE_VIEWPORT);
	addSignal("set-scroll-adjustments", &Gtk::Viewport::signal_set_scroll_adjustments);

	addingSignals(GTK_TYPE_NOTEBOOK);
	addSignal("switch-page", &Gtk::Notebook::signal_switch_page);
	addSignal("page-added", &Gtk::Notebook::signal_page_added);
	addSignal("page-removed", &Gtk::Notebook::signal_page_removed);
	addSignal("page-reordered", &Gtk::Notebook::signal_page_reordered);

	addingSignals(GTK_TYPE_ADJUSTMENT);
	addSignal("changed", &Gtk::Adjustment::signal_changed);
	addSignal("value-changed", &Gtk::Adjustment::signal_value_changed);

	addingSignals(GTK_TYPE_WINDOW);
	addSignal("frame-event", &Gtk::Window::signal_frame_event);
	addSignal("set-focus", &Gtk::Window::signal_set_focus);

	addingSignals(GTK_TYPE_ASSISTANT);
	addSignal("apply", &Gtk::Assistant::signal_apply);
	addSignal("cancel", &Gtk::Assistant::signal_cancel);
	addSignal("close", &Gtk::Assistant::signal_close);
	addSignal("prepare", &Gtk::Assistant::signal_prepare);

	addingSignals(GTK_TYPE_DIALOG);
	addSignal("response", &Gtk::Dialog::signal_response);

	addingSignals(GTK_TYPE_INFO_BAR);
	addSignal("response", &Gtk::InfoBar::signal_response);

	addingSignals(GTK_TYPE_FILE_CHOOSER);
	addSignal("current-folder-changed", &Gtk::FileChooser::signal_current_folder_changed);
	addSignal("file-activated", &Gtk::FileChooser::signal_file_activated);
	addSignal("selection-changed", &Gtk::FileChooser::signal_selection_changed);
	addSignal("update-preview", &Gtk::FileChooser::signal_update_preview);
	addSignal("confirm-overwrite", &Gtk::FileChooser::signal_confirm_overwrite);

	addingSignals(GTK_TYPE_FILE_CHOOSER_BUTTON);
	addSignal("file-set", &Gtk::FileChooserButton::signal_file_set);

	addingSignals(GTK_TYPE_STATUSBAR);
	addSignal("text-popped", &Gtk::Statusbar::signal_text_popped);
	addSignal("text-pushed", &Gtk::Statusbar::signal_text_pushed);

	addingSignals(GTK_TYPE_ACTION_GROUP);
	addSignal("connect-proxy", &Gtk::ActionGroup::signal_connect_proxy);
	addSignal("disconnect-proxy", &Gtk::ActionGroup::signal_disconnect_proxy);
	addSignal("post-activate", &Gtk::ActionGroup::signal_post_activate);
	addSignal("pre-activate", &Gtk::ActionGroup::signal_pre_activate);

	addingSignals(GTK_TYPE_ACTION);
	addSignal("activate", &Gtk::Action::signal_activate);

	addingSignals(GTK_TYPE_TOGGLE_ACTION);
	addSignal("toggled", &Gtk::ToggleAction::signal_toggled);

	addingSignals(GTK_TYPE_RADIO_ACTION);
	addSignal("changed", &Gtk::RadioAction::signal_changed);

	addingSignals(GTK_TYPE_UI_MANAGER);
	addSignal("actions-changed", &Gtk::UIManager::signal_actions_changed);
	addSignal("add-widget", &Gtk::UIManager::signal_add_widget);
	addSignal("connect-proxy", &Gtk::UIManager::signal_connect_proxy);
	addSignal("disconnect-proxy", &Gtk::UIManager::signal_disconnect_proxy);
	addSignal("post-activate", &Gtk::UIManager::signal_post_activate);
	addSignal("pre-activate", &Gtk::UIManager::signal_pre_activate);

	addingSignals(GTK_TYPE_MENU_SHELL);
	addSignal("deactivate", &Gtk::MenuShell::signal_deactivate);
	addSignal("selection-done", &Gtk::MenuShell::signal_selection_done);
	addSignal("move-selected", &Gtk::MenuShell::signal_move_selected);

	addingSignals(GTK_TYPE_TOOLBAR);
	addSignal("orientation-changed", &Gtk::Toolbar::signal_orientation_changed);
	addSignal("popup-context-menu", &Gtk::Toolbar::signal_popup_context_menu);
	addSignal("style-changed", &Gtk::Toolbar::signal_toolbar_style_changed);

	addingSignals(GTK_TYPE_LABEL);
	addSignal("populate-popup", &Gtk::Label::signal_populate_popup);

	addingSignals(GTK_TYPE_BUTTON);
	addSignal("activate", &Gtk::Button::signal_activate);
	addSignal("clicked", &Gtk::Button::signal_clicked);
	addSignal("enter", &Gtk::Button::signal_enter);
	addSignal("leave", &Gtk::Button::signal_leave);
	addSignal("pressed", &Gtk::Button::signal_pressed);
	addSignal("released", &Gtk::Button::signal_released);

	addingSignals(GTK_TYPE_TOGGLE_BUTTON);
	addSignal("toggled", &Gtk::ToggleButton::signal_toggled);

	addingSignals(GTK_TYPE_RADIO_BUTTON);
	addSignal("group-changed", &Gtk::RadioButton::signal_group_changed);

	addingSignals(GTK_TYPE_ENTRY_COMPLETION);
	addSignal("action-activated", &Gtk::EntryCompletion::signal_action_activated);
	addSignal("insert-prefix", &Gtk::EntryCompletion::signal_insert_prefix);
	addSignal("match-selected", &Gtk::EntryCompletion::signal_match_selected);
	addSignal("cursor-on-match", &Gtk::EntryCompletion::signal_cursor_on_match);

	addingSignals(GTK_TYPE_EDITABLE);
	addSignal("changed", &Gtk::Editable::signal_changed);
	addSignal("delete-text", &Gtk::Editable::signal_delete_text);
	addSignal("insert-text", &Gtk::Editable::signal_insert_text);

	addingSignals(GTK_TYPE_ENTRY);
	addSignal("activate", &Gtk::Entry::signal_activate);
	addSignal("insert-at-cursor", &Gtk::Entry::signal_insert_at_cursor);
	addSignal("populate-popup", &Gtk::Entry::signal_populate_popup);
	addSignal("icon-press", &Gtk::Entry::signal_icon_press);
	addSignal("icon-release", &Gtk::Entry::signal_icon_release);

	addingSignals(GTK_TYPE_COMBO_BOX);
	addSignal("changed", &Gtk::ComboBox::signal_changed);

	addingSignals(GTK_TYPE_SPIN_BUTTON);
	addSignal("input", &Gtk::SpinButton::signal_input);
	addSignal("output", &Gtk::SpinButton::signal_output);
	addSignal("value-changed", &Gtk::SpinButton::signal_value_changed);
	addSignal("wrapped", &Gtk::SpinButton::signal_wrapped);

	addingSignals(GTK_TYPE_RANGE);
	addSignal("adjust-bounds", &Gtk::Range::signal_adjust_bounds);
	addSignal("change-value", &Gtk::Range::signal_change_value);
	addSignal("move-slider", &Gtk::Range::signal_move_slider);
	addSignal("value-changed", &Gtk::Range::signal_value_changed);

	addingSignals(GTK_TYPE_SCALE);
	addSignal("format-value", &Gtk::Scale::signal_format_value);

	addingSignals(GTK_TYPE_TEXT_VIEW);
	addSignal("insert-at-cursor", &Gtk::TextView::signal_insert_at_cursor);
	addSignal("populate-popup", &Gtk::TextView::signal_populate_popup);
	addSignal("set-anchor", &Gtk::TextView::signal_set_anchor);
	addSignal("set-scroll-adjustments", &Gtk::TextView::signal_set_scroll_adjustments);

	addingSignals(GTK_TYPE_ICON_VIEW);
	addSignal("item-activated", &Gtk::IconView::signal_item_activated);
	addSignal("selection-changed", &Gtk::IconView::signal_selection_changed);
	addSignal("set-scroll-adjustments", &Gtk::IconView::signal_set_scroll_adjustments);

	addingSignals(GTK_TYPE_TREE_VIEW);
	addSignal("columns-changed", &Gtk::TreeView::signal_columns_changed);
	addSignal("cursor-changed", &Gtk::TreeView::signal_cursor_changed);
	addSignal("row-activated", &Gtk::TreeView::signal_row_activated);
	addSignal("row-collapsed", &Gtk::TreeView::signal_row_collapsed);
	addSignal("row-expanded", &Gtk::TreeView::signal_row_expanded);
	addSignal("set-scroll-adjustments", &Gtk::TreeView::signal_set_scroll_adjustments);
	addSignal("test-collapse-row", &Gtk::TreeView::signal_test_collapse_row);
	addSignal("test-expand-row", &Gtk::TreeView::signal_test_expand_row);

	addingSignals(GTK_TYPE_COLOR_BUTTON);
	addSignal("color-set", &Gtk::ColorButton::signal_color_set);

	addingSignals(GTK_TYPE_COLOR_SELECTION);
	addSignal("color-changed", &Gtk::ColorSelection::signal_color_changed);

	addingSignals(GTK_TYPE_FONT_BUTTON);
	addSignal("font-set", &Gtk::FontButton::signal_font_set);

	addingSignals(GTK_TYPE_LAYOUT);
	addSignal("set-scroll-adjustments", &Gtk::Layout::signal_set_scroll_adjustments);

	addingSignals(GTK_TYPE_CALENDAR);
	addSignal("day-selected", &Gtk::Calendar::signal_day_selected);
	addSignal("day-selected-double-click", &Gtk::Calendar::signal_day_selected_double_click);
	addSignal("month-changed", &Gtk::Calendar::signal_month_changed);
	addSignal("next-month", &Gtk::Calendar::signal_next_month);
	addSignal("next-year", &Gtk::Calendar::signal_next_year);
	addSignal("prev-month", &Gtk::Calendar::signal_prev_month);
	addSignal("prev-year", &Gtk::Calendar::signal_prev_year);

	addingSignals(GTK_TYPE_STATUS_ICON);
	addSignal("size-changed", &Gtk::StatusIcon::signal_size_changed);
	addSignal("activate", &Gtk::StatusIcon::signal_activate);
	addSignal("popup-menu", &Gtk::StatusIcon::signal_popup_menu);
	addSignal("button-press-event", &Gtk::StatusIcon::signal_button_press_event);
	addSignal("button-release-event", &Gtk::StatusIcon::signal_button_release_event);
	addSignal("scroll-event", &Gtk::StatusIcon::signal_scroll_event);
	addSignal("query-tooltip", &Gtk::StatusIcon::signal_query_tooltip);

	addingSignals(GTK_TYPE_RECENT_CHOOSER);
	addSignal("item-activated", &Gtk::RecentChooser::signal_item_activated);
	addSignal("selection-changed", &Gtk::RecentChooser::signal_selection_changed);

	addingSignals(GTK_TYPE_SCALE_BUTTON);
	addSignal("value-changed", &Gtk::ScaleButton::signal_value_changed);

	addingSignals(G_TYPE_INVALID);
}

LoaderInit::~LoaderInit() {
	for(Signals::iterator it= signals.begin(); it!=signals.end(); ++it)
		delete it->second.second;
}

}
